/* This file is part of Idest.
   Copyright (C) 2009-2011 Sergey Poznyakoff
 
   Idest is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.
 
   Idest is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with Idest.  If not, see <http://www.gnu.org/licenses/>. */

#include "idest.h"

unsigned version_option = 2;
unsigned convert_version;
int latin1_option = 0;
enum backup_type backup_type = no_backups;
char *backup_dir;
int verbose_option = 0;

int all_frames = 0;
gl_list_t ed_list;

static bool
ed_item_eq(const void *elt1, const void *elt2)
{
	const struct ed_item *i1 = elt1;
	const struct ed_item *i2 = elt2;
	return strcmp(i1->id, i2->id) == 0;
}

static void
ed_item_dispose(const void *elt)
{
	free((void*)elt);
}

void
ed_list_create()
{
	if (ed_list)
		return;
	ed_list = gl_list_create_empty(&gl_linked_list_implementation,
				       ed_item_eq,
				       NULL,
				       ed_item_dispose,
				       false);
}

struct ed_item *
ed_item_create0(char *name, const char *id)
{
	struct ed_item *itm = xzalloc(sizeof(itm[0]));
	itm->name = name;
	strncpy(itm->id, id, sizeof(itm->id));
	return itm;
}

struct ed_item *
ed_item_create(const char *name, const char *id)
{
	return ed_item_create0(xstrdup(name), id);
}


int item_ids[] = {
	item_title,
	item_artist,
	item_album,
	item_year,
	item_comment,
	item_track,
	item_genre
};

const char *item_names[] = {
	"title",
	"artist",
	"album",
	"year",
	"comment",
	"track",
	"genre",
	NULL
};
ARGMATCH_VERIFY(item_names, item_ids);

const char *item_frames[] = {
	ID3_FRAME_TITLE,
	ID3_FRAME_ARTIST,
	ID3_FRAME_ALBUM,
	ID3_FRAME_YEAR,
	ID3_FRAME_COMMENT,
	ID3_FRAME_TRACK,
	ID3_FRAME_GENRE,
	NULL
};

ARGMATCH_VERIFY(item_frames, item_ids);

int
item_id(const char *arg)
{
	return (int) ARGMATCH(arg, item_names, item_ids);
}

int
frame_id(const char *arg)
{
	return (int) ARGMATCH(arg, item_frames, item_ids);
}

static struct ed_item *
ed_list_new_item(const char *arg, size_t len)
{
	int d;
	char *name;
	const char *id;
	struct ed_item *itm;
	
	name = xmalloc(len+1);
	memcpy(name, arg, len);
	name[len] = 0;

	d = item_id(name);
	if (d < 0) {
		if (!id3_frametype_lookup(name, len)) 
			error(1, 0, d == -1
			              ? "%s: unknown frame name" :
			                "%s: ambiguous frame name",
			      name);
		id = name;
	} else {
		if (strcmp(item_names[d], name)) {
			free(name);
			name = xstrdup(item_names[d]);
		}
		id = item_frames[d];
	}
	itm = ed_item_create0(name, id);
	gl_list_add_last(ed_list, itm);
	return itm;
}

void
parse_ed_items(const char *arg)
{
	ed_list_create();
	while (*arg) {
		size_t len = strcspn(arg, " \t,");

		ed_list_new_item(arg, len);

		arg += len;
		if (!*arg)
			break;
		arg += strspn(arg, " \t,");
	}
}

void
ed_list_add_item(const char *id, gl_list_t list)
{
	if (all_frames) {
		const char *name;
		struct ed_item *itm;
		int i;
		
		ed_list_create();
		i = frame_id(id);
		if (i >= 0)
			name = item_names[i];
		else
			name = id;
		itm = ed_item_create(name, id);
		itm->v.vlist = list;
		gl_list_add_last(ed_list, itm);
	} else {
		gl_list_node_t node;
		struct ed_item item;
		strncpy(item.id, id, sizeof(item.id));
		node = gl_list_search(ed_list, &item);
		if (node) {
			struct ed_item *itm =
				(struct ed_item *) gl_list_node_value(ed_list,
								      node);
			if (itm->v.vlist) {
				concat_string_list(itm->v.vlist, list);
				gl_list_free(list);
			} else 
				itm->v.vlist = list;
		} else
			gl_list_free(list);
	}
}

void
ed_list_add_assignment(const char *name, const char *value)
{
	struct ed_item *itm;

	ed_list_create();
	itm = ed_list_new_item(name, strlen(name));
	itm->v.value = xstrdup(value);
}

void
ed_list_print()
{
	gl_list_iterator_t itr;
	const void *p;
	if (!ed_list)
		return;
	itr = gl_list_iterator(ed_list);
	while (gl_list_iterator_next(&itr, &p, NULL)) {
		const struct ed_item *item = p;
		printf("%s:", item->name);
		if (item->v.vlist) 
			print_string_list(stdout, item->v.vlist);
		putchar('\n');
	}
	gl_list_iterator_free(&itr);
}

void
ed_list_clear()
{
	gl_list_iterator_t itr;
	const void *p;
	if (!ed_list)
		return;
	itr = gl_list_iterator(ed_list);
	while (gl_list_iterator_next(&itr, &p, NULL)) {
		struct ed_item *item = (struct ed_item *)p;
		if (item->v.vlist) {
			gl_list_free(item->v.vlist);
			item->v.vlist = NULL;
		}
	}
	gl_list_iterator_free(&itr);
}

const struct ed_item *
ed_item_find(int id)
{
	if (ed_list) {
		gl_list_node_t node;
		struct ed_item item;
		strncpy(item.id, item_frames[id], sizeof(item.id));
		node = gl_list_search(ed_list, &item);
		if (node) 
			return gl_list_node_value(ed_list, node);
	}
	return NULL;
}


void
verify_mp3(FILE *fp, const char *name)
{
	char str[3];
	unsigned short sig = 0;

	if (fread(str, sizeof(str), 1, fp) != 1) 
		error(1, errno, "error reading %s", name);
	if (memcmp(str, "ID3", 3) == 0) {
		error(0, 0, "%s: warning: ID3 header found",
		      name);
		return;
	}
	sig = ((unsigned char)str[0]) << 8 | ((unsigned char)str[1]);

	switch (sig & 0xFFFE) {
	case 0xFFFA: /* MPEG ADTS, layer III, v1 */
	case 0xFFF2: /* MPEG ADTS, layer III, v2 */
	case 0xFFFE: /* MPEG ADTS, layer III, v2.5 */
		break;

	default:
		error(1, 0, "%s: not an mp3 file", name);
	}
}

void
query_id3(const char *name)
{
	if (all_frames)
		ed_list = NULL;
	else if (!ed_list)
		parse_ed_items(DEFAULT_ED_LIST);
	query_tags(name);
}

void
set_id3(const char *name)
{
	if (backup_file(name)) 
		error(1, 0, "cannot backup file %s", name);
	set_tags(name);
}

void
del_id3(const char *name)
{
	if (backup_file(name)) 
		error(1, 0, "cannot backup file %s", name);
	del_tags(name);
}

void (*id3_mode[])(const char *) = {
	query_id3,
	set_id3,
	del_id3,
};

int mode = MODE_QUERY;

#include "cmdline.h"

int
main(int argc, char **argv)
{
	set_program_name(argv[0]);

	get_options(argc, argv);
		
	argc -= optind;
	argv += optind;

	if (argc == 0) 
		error(1, 0, "no files");

	guile_init();

	if (convert_version)
		mode = MODE_MOD;

	while (argc--) {
		char *name = *argv++;
		if (verbose_option)
			printf("%s:\n", name);
		id3_mode[mode](name);
	}
	
	exit(0);
}
